# SPDX-FileCopyrightText: 2025 Florian Obersteiner <f.obersteiner@kit.edu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging as log
from datetime import datetime
from pathlib import Path
from typing import Iterable

import pandas as pd
import xarray as xr
from siphon.catalog import Dataset, TDSCatalog
from siphon.catalog import Dataset as SiphonDataset

from .config import DotDict


def get_ms_datasets(cfg: DotDict) -> list[Dataset]:
    """Obtain references to measurement data from THREDDS"""

    log.info("begin query THREDDS...")

    catalog = TDSCatalog(cfg.THREDDS_URL)
    flights = (
        list(range(cfg.FLIGHTNO_MIN, cfg.FLIGHTNO_MAX + 1))
        if cfg.FLIGHTNO_MIN != -1 and cfg.FLIGHTNO_MAX != -1
        else cfg.FLIGHTS
    )
    datasets = []

    for i, ds in enumerate(catalog.datasets):
        if not (ds.startswith(cfg.FNAME_PREFIX) and ds.endswith(cfg.FNAME_SUFFIX)):
            continue
        if int(ds.split("_")[2]) not in flights:
            continue

        datasets.append(catalog.datasets[i])

    log.info("THREDDS query completed")

    return datasets


def ms_from_thredds(
    datasets: list[SiphonDataset], config: DotDict, apconfig: pd.DataFrame, drop: bool = False
) -> (pd.DataFrame, dict[int, str]):  # type: ignore
    """
    Load multiple netCDF files from a THREDS TDS into a single DataFrame.

    Raises:
        ValueError: If 'ACN' not found in dataset.
        ValueError: If no datasets are found.
    """

    log.info("begin load data from THREDDS...")

    dataframes = []
    airports = {}

    for i, ds in enumerate(datasets):
        log.info(f"load {ds.name}")
        fullset = datasets[i].remote_access(use_xarray=True).to_dataframe()  # type: ignore
        if "ACN" in fullset.columns and fullset["ACN"].isna().all():  # type: ignore
            log.warning(f"'ACN' column is all NaN values! Skip {ds.name}")
            continue

        if drop:
            df = pd.DataFrame()
            for v in config.VARIABLES:
                df[v] = fullset[v]
        else:
            df = fullset

        # extract flight number and date from filename; add as extra columns
        df["flight_number"] = int(ds.name.split("_")[2])
        airports[int(ds.name.split("_")[2])] = (
            f"from: {ds.name.split('_')[3]} ({apconfig.loc[ds.name.split('_')[3], :].Name}), to: {ds.name.split('_')[4]} ({apconfig.loc[ds.name.split('_')[4], :].Name})"
        )

        df["date"] = datetime.strptime(ds.name.split("_")[1], "%Y%m%d").strftime("%Y-%m-%d")
        # make a linear interpolation of the acetonitrile column for BB flaggin
        if "ACN" in df.columns:
            df["BB_flag"] = df["ACN"].interpolate() > (
                config.BB_THRESH + config.ACN_PRC_NSIGMA * df["ACN_prc"].interpolate()
            )
        else:
            raise ValueError(f"ACN column not found in {ds.name}")

        dataframes.append(df)

    if len(dataframes) == 0:
        raise ValueError("No datasets found!")

    log.info("data loaded from THREDDS successfully")

    return (pd.concat(dataframes), airports)


def ms_from_disk(
    paths: Iterable[Path], config: DotDict, apconfig: pd.DataFrame, drop: bool = False
) -> (pd.DataFrame, dict[int, str]):  # type: ignore
    """
    Load multiple netCDF files into a single DataFrame.

    Raises:
        ValueError: If 'ACN' not found in dataset.
        ValueError: If no datasets are found.
    """

    log.info("begin load data from disk...")

    dataframes = []
    airports = {}

    for path in sorted(paths):
        fullset = xr.open_dataset(path).to_dataframe()
        if "ACN" in fullset.columns and fullset["ACN"].isna().all():  # type: ignore
            log.warning(f"'ACN' column is all NaN values! Skip {path.name}")
            continue

        if drop:
            df = pd.DataFrame()
            for v in config.VARIABLES:
                df[v] = fullset[v]
        else:
            df = fullset

        # extract flight number and date from filename; add as extra columns
        df["flight_number"] = int(path.name.split("_")[2])
        airports[int(path.name.split("_")[2])] = (
            f"from: {path.name.split('_')[3]} ({apconfig.loc[path.name.split('_')[3], :].Name}), to: {path.name.split('_')[4]} ({apconfig.loc[path.name.split('_')[4], :].Name})"
        )

        df["date"] = datetime.strptime(path.stem.split("_")[1], "%Y%m%d").strftime("%Y-%m-%d")
        # make a linear interpolation of the acetonitrile column for BB flaggin
        if "ACN" in df.columns:
            if df["ACN"].isna().all():  # type: ignore
                log.warning("'ACN' column is all NaN values!")
            df["BB_flag"] = df["ACN"].interpolate() > (
                config.BB_THRESH + config.ACN_PRC_NSIGMA * df["ACN_prc"].interpolate()
            )
        else:
            raise ValueError(f"ACN column not found in {path}")

        dataframes.append(df)

    if len(dataframes) == 0:
        raise ValueError("no datasets found!")

    log.info("data loaded from disk successfully")

    return (pd.concat(dataframes), airports)


def tr_from_disk(paths: Iterable[Path]) -> dict[int, xr.Dataset]:

    log.info("begin load trajectory data from disk...")

    d = {}
    for p in paths:
        fno = int(p.name.split("_")[-2])
        assert d.get(fno) is None  # assert there are no dupes
        d[fno] = xr.load_dataset(p)

    log.info("trajectory data loaded from disk successfully")

    return d


def tr_from_thredds(cfg: DotDict) -> dict[int, xr.Dataset]:
    """
    Load multiple netCDF files from a THREDS TDS into a dict,
    mapping flight number -> xarray.Dataset.

    Raises:
        ValueError: If no datasets are found.
    """

    log.info("begin load trajectory data from THREDDS...")

    d = {}

    catalog = TDSCatalog(cfg.THREDDS_URL_TRDATA)
    flights = (
        list(range(cfg.FLIGHTNO_MIN, cfg.FLIGHTNO_MAX + 1))
        if cfg.FLIGHTNO_MIN != -1 and cfg.FLIGHTNO_MAX != -1
        else cfg.FLIGHTS
    )

    for i, ds in enumerate(catalog.datasets):  # 'ds' is treated as string here
        if not (ds.startswith(cfg.FNAME_PREFIX_TR) and ds.endswith(cfg.FNAME_SUFFIX)):
            log.warning(f"invalid dataset '{ds}'")
            continue
        fno = int(ds.split("_")[4])
        if fno not in flights:
            log.warning(f"dataset '{ds}' not in specified flight range")
            continue

        log.info(f"load {ds}")

        assert d.get(fno) is None  # assert there are no dupes
        d[fno] = catalog.datasets[i].remote_access(use_xarray=True)

    log.info("trajectory data loaded from THREDDS successfully")

    return d
